#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sysexits.h>
#include <errno.h>
#include <syslog.h>

#include <sys/ioctl.h>
#include <sys/dumbups.h>
#include <sys/reboot.h>
#include <sys/lkm.h>
#include <signal.h>

/*
 * Determine ldisc number for dbups
 */

static int getldisc(void)
{
	struct lmc_stat sbuf;
	int fd;

	if ((fd = open ("/dev/lkm", O_RDONLY, 0)) < 0) {
		perror ("open(\"/dev/lkm\")");
		return -1;
	}
	bzero (&sbuf, sizeof (sbuf));
	strcpy (sbuf.name, "dbups_mod");
	if (ioctl (fd, LMSTAT, &sbuf) < 0) {
		perror ("ioctl(LMSTAT)");
		return -1;
	}
	return sbuf.offset;
}


int print_modem (int fd)
{
	int i;

	if (ioctl (fd, TIOCMGET, &i) < 0) {
		perror ("ioctl(TIOCMGET)");
		return EX_OSERR;
	}
	printf ("%s %s | %s %s %s (%d)\r",
		i & TIOCM_DTR ? "+DTR" : "-dtr",
		i & TIOCM_RTS ? "+RTS" : "-rts",
		i & TIOCM_DSR ? "+DSR" : "-dsr",
		i & TIOCM_CTS ? "+CTS" : "-cts",
		i & TIOCM_CD  ? "+CD " : "-cd ",
		i);
	fflush (stdout);
	return 0;
}


static void doreboot (void)
{
	int i;

	/* Just stop init -- if we fail, we'll restart it. */
	if (kill(1, SIGTSTP) == -1)
		perror("SIGTSTP init");

	/* Ignore the SIGHUP we get when our parent shell dies. */
	(void)signal(SIGHUP, SIG_IGN);

	/* Send a SIGTERM first, a chance to save the buffers. */
	if (kill(-1, SIGTERM) == -1)
		perror("SIGTERM processes");

	/*
	 * After the processes receive the signal, start the rest of the
	 * buffers on their way.  Wait 5 seconds between the SIGTERM and
	 * the SIGKILL to give everybody a chance.
	 */
	sleep(2);
	sync();
	sleep(3);

	for (i = 1;; ++i) {
		if (kill(-1, SIGKILL) == -1) {
			if (errno == ESRCH)
				break;
			break;
		}
		if (i > 5) {
			(void)fprintf(stderr,
			    "WARNING: some process(es) wouldn't die\n");
			break;
		}
		(void)sleep(2 * i);
	}

	reboot (RB_POWEROFF);
}


int main (void)
{
	int fd, i;
	struct ups_states us, us2;
	int failing, battfailing;

	if ((fd=open("/dev/ttyd1", O_RDWR, 0)) < 0) {
		perror ("open");
		return EX_OSERR;
	}

	i = 0;
	ioctl(fd, TIOCMSET, &i);

	i = getldisc();
	if (i < 0) {
		return EX_OSERR;
	}
	if (ioctl (fd, TIOCSETD, &i) < 0) {
		perror ("ioctl(TIOCSETD)");
		return EX_OSERR;
	}

#ifdef DEBUG
	printf ("Before: ");
	print_modem(fd);
	puts("");
#endif

	us.ups_init.upss_sm = TIOCM_DTR;
	us.ups_init.upss_cm = TIOCM_RTS;
	us.ups_off.upss_sm  = TIOCM_RTS;
	us.ups_off.upss_cm  = 0;
	us.ups_linefail.upss_sm = 0;
	us.ups_linefail.upss_cm = TIOCM_CTS;
	us.ups_battfail.upss_sm = 0;
	us.ups_battfail.upss_cm = TIOCM_CTS;

	if (ioctl (fd, BIOSSTMASK, &us) < 0) {
		perror ("ioctl(BIOSSTMASK)");
		return EX_OSERR;
	}

	if (ioctl (fd, BIOGSTMASK, &us2) < 0) {
		perror ("ioctl(BIOGSTMASK)");
		return EX_OSERR;
	}
	
	if (bcmp (&us, &us2, sizeof(us)) != 0) {
		printf ("set failed.\n");
		return EX_OSERR;
	}

	openlog ("upsd", LOG_DAEMON, 0);
#if !defined(DEBUG)
	daemon (0, 0);
#endif

	failing = battfailing = 0;

	while (1) {
		sleep (1);
		if (ioctl (fd, BIOGSTATUS, &i) < 0) {
    		perror ("ioctl(BIOGSTATE)");
    		return EX_OSERR;
    	}
#ifdef DEBUG
		printf ("line %s,batt %s, ", 
			i & UPSSTATUS_LINEFAIL ? "FAIL" : "OK  ",
			i & UPSSTATUS_BATTFAIL ? "FAIL" : "OK  ");
		print_modem(fd);
#endif
		if (i & UPSSTATUS_LINEFAIL) {
			if (!failing)
				syslog (LOG_NOTICE, "UPS reports power line failure");
			failing = 1;
		} else {
			if (failing) {
				syslog (LOG_NOTICE, "UPS reports line power restored");
				failing = 0;
			}
		}

		if (i & UPSSTATUS_BATTFAIL) {
			syslog (LOG_EMERG, "UPS reports battery failure");
			sleep (5);
			if (ioctl (fd, BIOGSTATUS, &i) < 0) {
    			perror ("ioctl(BIOGSTATE)");
    			return EX_OSERR;
	    	}
    		if (i & UPSSTATUS_BATTFAIL) {
    			syslog (LOG_EMERG, "UPS reports battery failure, shutting down system");
    			doreboot();
			} else {
    			syslog (LOG_EMERG, "UPS reports battery restored");
			}
		}
	}
	
	return 0;

}

